"""
Adapted from openmoltools.openeye.smiles_to_antechamber(...)
by Caitlin C. Bannan March 2016

I added the variable storeConf so that it will return a generated conformation other than the first. 
Below the section of code that has been editted is in the #====== section. 
"""

import shutil
import os
import mdtraj as md
from openmoltools.utils import import_, enter_temp_directory, run_antechamber, create_ffxml_file
import logging
from openmoltools.openeye import *

def smiles_to_antechamber_variableConformations(smiles_string, gaff_mol2_filename, frcmod_filename, residue_name="MOL", strictStereo=False, storeConf = 1):
    """
    This is a variation on the openmoltools.openeye.smiles_to_antechamber 
    The only difference is it returns the nth generated conformation where
    n is given by storeConf, the default would behave the same as the original
    method returning the first generated conformation
    
    Build a molecule from a smiles string and run antechamber, 
    generating GAFF mol2 and frcmod files from a smiles string.  Charges
    will be generated using the OpenEye QuacPac AM1-BCC implementation.
    
    Parameters
    ----------
    smiles_string : str
        Smiles string of molecule to construct and charge
    gaff_mol2_filename : str
        Filename of mol2 file output of antechamber, with charges
        created from openeye
    frcmod_filename : str
        Filename of frcmod file output of antechamber.  Most likely
        this file will be almost empty, at least for typical molecules.
    residue_name : str, optional, default="MOL"
        OpenEye writes mol2 files with <0> as the residue / ligand name.
        This chokes many mol2 parsers, so we replace it with a string of
        your choosing.  This might be useful for downstream applications
        if the residue names are required to be unique.
    strictStereo : bool, optional, default=False
        If False, permits smiles strings with unspecified stereochemistry.
        See https://docs.eyesopen.com/omega/usage.html
    """
    oechem = import_("openeye.oechem")
    if not oechem.OEChemIsLicensed(): raise(ImportError("Need License for oechem!"))    

    # Get the absolute path so we can find these filenames from inside a temporary directory.
    gaff_mol2_filename = os.path.abspath(gaff_mol2_filename)
    frcmod_filename = os.path.abspath(frcmod_filename)
    
    m = smiles_to_oemol(smiles_string)
    #=============================================================================================================================
    # CCB: This is the part I changed so that we can get a different than the first conformer.    
    # This will return all generated conformations
    m = get_charges(m, strictStereo=strictStereo, keep_confs=-1) # In the original keep_confs = 1 

    # The next 5 lines of code are completely new. 
    # Check that the number of conformations is at least as big as storeConf 
    if m.NumConfs() < storeConf:
        raise Exception("storeConf %i is bigger than the number of available conformations (%i)" % (storeConf, m.NumConfs()))
    # Now only save the storeConf th conformation
    for k, conf in enumerate(m.GetConfs() ):
        if k != (storeConf - 1):
            m.DeleteConf(conf)
    #=============================================================================================================================

    with enter_temp_directory():  # Avoid dumping 50 antechamber files in local directory.    
        _unused = molecule_to_mol2(m, "./tmp.mol2", residue_name=residue_name)
        net_charge = oechem.OENetCharge(m)
        tmp_gaff_mol2_filename, tmp_frcmod_filename = run_antechamber("tmp", "./tmp.mol2", charge_method=None, net_charge=net_charge)  # USE OE AM1BCC charges!        
        shutil.copy(tmp_gaff_mol2_filename, gaff_mol2_filename)
        shutil.copy(tmp_frcmod_filename, frcmod_filename)
